defmodule Clickhouse.Interface.HTTP do
  @moduledoc """
  An interface to interact with a Clickhouse server via HTTP.
  """

  @behaviour Clickhouse.Interface

  use GenServer

  require Logger

  alias Clickhouse.{Error, Format, Query, Result}

  @opts_schema %{
    client: [type: :atom, required: true, default: :default],
    urls: [type: {:list, :binary}, required: true, default: ["http://localhost:8123"]],
    connect_retry: [type: :integer, required: true, default: 3_000]
  }

  @http_opts_schema %{
    timeout: [type: :timeout, required: false, default: 5_000],
    transport_opts: [type: :list, required: false, default: []]
  }

  @ping_statement "SELECT 1"

  ################################
  # Clickhouse.Interface Callbacks
  ################################

  @doc """
  Starts the HTTP interface server.
  """
  @impl Clickhouse.Interface
  @spec start_link(keyword()) :: GenServer.on_start()
  def start_link(opts) do
    with {:ok, opts} <- validate_opts(opts, @opts_schema) do
      GenServer.start_link(__MODULE__, opts)
    end
  end

  @doc """
  Executes a query via HTTP.
  """
  @impl Clickhouse.Interface
  @spec execute(Clickhouse.Query.t()) ::
          {:error, Clickhouse.Error.t()} | {:ok, Clickhouse.Result.t()}
  def execute(query) do
    with {:ok, urls} <- config(query.client, :urls),
         {:ok, opts} <- validate_opts(query.opts, @http_opts_schema) do
      statement = Query.get_statement(query)
      do_execute(urls, statement, opts)
    end
  end

  ################################
  # GenServer Callbacks
  ################################

  @impl GenServer
  def init(opts) do
    with {:ok, opts} <- validate_opts(opts, @opts_schema) do
      state = init_state(opts)
      init_config(state)
      {:ok, state, {:continue, :connect}}
    end
  end

  @impl GenServer
  def handle_continue(:connect, state) do
    connect(state)
    {:noreply, state}
  end

  @impl GenServer
  def handle_info(:retry_connect, state) do
    {:noreply, state, {:continue, :connect}}
  end

  ################################
  # Private API
  ################################

  defp validate_opts(opts, schema) do
    case KeywordValidator.validate(opts, schema) do
      {:ok, _} = result -> result
      {:error, reason} -> {:error, Error.new(:argument_error, reason)}
    end
  end

  defp init_state(opts) do
    client = Keyword.get(opts, :client)

    %{
      client: client,
      table_name: table_name(client),
      urls: Keyword.get(opts, :urls),
      connect_retry: Keyword.get(opts, :connect_retry)
    }
  end

  defp init_config(state) do
    :ets.new(state.table_name, [:named_table, read_concurrency: true])
    :ets.insert(state.table_name, {:urls, state.urls})
  end

  defp table_name(client) do
    :"#{__MODULE__}.#{client}"
  end

  defp config(client, key) do
    [{^key, value}] = :ets.lookup(table_name(client), key)
    {:ok, value}
  rescue
    ArgumentError ->
      {:error, Error.new(:not_found, "HTTP interface #{inspect(client)} not found.")}
  end

  defp connect(state) do
    case do_execute(state.urls, @ping_statement, []) do
      {:ok, _} ->
        :ok

      {:error, error} ->
        Logger.error("[Clickhouse] Connection error: #{inspect(error)}")
        Process.send_after(self(), :retry_connect, state.connect_retry)
        :error
    end
  end

  defp do_execute([url], statement, opts) do
    do_execute(url, statement, opts)
  end

  defp do_execute(urls, statement, opts) when is_list(urls) do
    url = Enum.random(urls)
    do_execute(url, statement, opts)
  end

  defp do_execute(url, statement, opts) when is_binary(url) do
    case Mojito.post(url, [], statement, opts) do
      {:ok, %{status_code: 200} = response} ->
        build_result(response)

      {:ok, %{status_code: 400} = response} ->
        {:error, Error.new(:bad_request, response.body)}

      {:ok, %{status_code: 404} = response} ->
        {:error, Error.new(:not_found, response.body)}

      {:ok, %{status_code: code} = response} when code >= 500 ->
        {:error, Error.new(:server_error, response.body)}

      {:error, %{reason: %{reason: reason}}} ->
        {:error, Error.new(reason, nil)}

      {:error, %{reason: reason, message: message}} ->
        {:error, Error.new(reason, message)}
    end
  end

  defp build_result(response) do
    with {:ok, format} <- parse_format(response.headers) do
      result = Result.new(response.body, format)
      {:ok, result}
    end
  end

  defp parse_format([]), do: {:ok, nil}

  defp parse_format([{"x-clickhouse-format", format} | _headers]) do
    Format.find(format)
  end

  defp parse_format([_ | headers]) do
    parse_format(headers)
  end
end
